import logging
import os
import tempfile
import zipfile
from typing import Optional, List

from lxml import etree
from .file import read_epub_metadata
from .epub import (
    find_metadata_items,
    get_metadata_main, get_metadata_ext,
    get_metadata_node,
    get_book_sequence, set_book_sequence,
)


class Book:
    def __init__(self, fname: str, log: logging.Logger):
        self._file = fname
        self._log = log
        self._tree = read_epub_metadata(fname)
        self._metadata = None
        self._metadata = get_metadata_node(self._tree)
        series, num = get_book_sequence(self._tree, self._metadata)
        self._series = series
        self._series_num = num
        for el in self._metadata:
            if 'title' in el.tag:
                self._title = el
            if 'creator' in el.tag:
                self._creator = el
            if 'language' in el.tag:
                self._language = el
            if 'identifier' in el.tag:
                self._identifier = el

    @property
    def language(self) -> str:
        return self._language.text

    @property
    def title(self) -> str:
        return self._title.text

    @title.setter
    def title(self, title: str):
        self._title.text = title

    @property
    def creator(self) -> str:
        return self._creator.text

    @creator.setter
    def creator(self, creator: str):
        self._creator.text = creator

    @property
    def series(self) -> str:
        return self._series

    @series.setter
    def series(self, s: str) -> None:
        self._series = s
        set_book_sequence(self._tree, self._metadata, self._series, self._series_num)

    @property
    def series_num(self) -> int:
        return self._series_num

    @series_num.setter
    def series_num(self, n: int) -> None:
        self._series_num = n
        set_book_sequence(self._tree, self._metadata, self._series, self._series_num)

    def get_metadata(self, key: str, ns: str = None) -> List:
        return find_metadata_items(self._tree, self._metadata, key, ns)

    def get_dc(self, key: str) -> Optional[str]:
        return get_metadata_main(self._tree, self._metadata, key)

    def get_meta(self, key: str) -> Optional[str]:
        return get_metadata_ext(self._tree, self._metadata, key)

    def update(self):
        self._log.debug(f'file "{self._file}" updating')
        # create tmp file
        tmpfd, tmpname = tempfile.mkstemp(dir=os.path.dirname(self._file))
        os.close(tmpfd)
        with zipfile.ZipFile(self._file, 'r') as zin:
            with zipfile.ZipFile(tmpname, 'w') as zout:
                zout.comment = zin.comment  # preserve the comment
                for item in zin.infolist():
                    content = zin.read(item.filename)
                    if 'content.opf' in item.filename:
                        content = etree.tostring(self._tree,
                                                 method='xml', pretty_print=True,
                                                 encoding='utf-8', xml_declaration=False,
                                                 doctype='<?xml version="1.0" encoding="UTF-8"?>')
                    zout.writestr(item, content)
        # replace with the temp archive
        os.remove(self._file)
        os.rename(tmpname, self._file)

    def __repr__(self):
        return f"<Book(creator='{self.creator}', title='{self.title}')>"



import zipfile

from lxml import etree


def read_epub_metadata(f: str) -> etree:
    """read XML from 'content.opf' file in epub zip archive"""
    with zipfile.ZipFile(f, 'r') as zin:
        for item in zin.infolist():
            if 'content.opf' in item.filename:
                content = zin.read(item.filename)
                return etree.fromstring(content)



from typing import Tuple, Iterable, List, Optional

from lxml import etree

META_FB2_SEQ = 'FB2.book-info.sequence'
META_CALIBRE_SEQ = 'calibre:series'
META_CALIBRE_INX = 'calibre:series_index'


def get_meta_item(root: etree, metadata: Iterable[object], key: str) -> List:
    """look for `key` in meta elements"""
    els = find_metadata_items(root, metadata, 'meta', None)
    return [el for el in els if key == el.attrib['name']]


def get_metadata_ext(root: etree, metadata: Iterable[object], key: str) -> Optional[str]:
    """look for `key` in meta elements and returns value if `content` attribute"""
    els = get_meta_item(root, metadata, key)
    for el in els:
        return el.attrib['content']
    return None


def get_metadata_main(root: etree, metadata: Iterable[object], key: str) -> Optional[str]:
    """find nodes in main namespace and return content of the first element"""
    els = find_metadata_items(root, metadata, key, 'dc')
    return els[0].text if len(els) > 0 else None


def find_metadata_items(root: etree, metadata: Iterable[object], key: str, ns: str = None) -> List:
    """return list of metadata(nodes) in provided namespace"""
    if ns not in root.nsmap:
        return []
    fkey = f'{{{root.nsmap[ns]}}}{key}'
    return [el for el in metadata if fkey == el.tag]


def get_metadata_node(root: etree) -> Iterable[object]:
    for el in root:
        if 'metadata' in el.tag:
            return el


def get_book_sequence(root: etree, metadata: Iterable[object]) -> Tuple[str, int]:
    """parse FB2 sequence metadata"""
    series, num = '', -1
    # try read calibre metadata
    content = get_metadata_ext(root, metadata, META_CALIBRE_SEQ)
    if content is not None:
        series = content
    content = get_metadata_ext(root, metadata, META_CALIBRE_INX)
    if content is not None:
        num = int(content)
    if num > -1:
        return series, num
    # try read fb2
    content = get_metadata_ext(root, metadata, META_FB2_SEQ)
    if content is not None:
        arr = content.split(';')
        if len(arr) == 2:
            series = arr[0]
            num = int(''.join(filter(str.isdigit, arr[1])))
            return series, num
    return series, num


def remove_book_sequence(root: etree, metadata: Iterable[object]) -> None:
    """remove sequence metadata from book xml"""
    items = get_meta_item(root, metadata, META_FB2_SEQ)
    if len(items) > 0:
        content = items[0]
        content.getparent().remove(content)
    items = get_meta_item(root, metadata, META_CALIBRE_SEQ)
    if len(items) > 0:
        content = items[0]
        content.getparent().remove(content)
    items = get_meta_item(root, metadata, META_CALIBRE_INX)
    if len(items) > 0:
        content = items[0]
        content.getparent().remove(content)


def set_book_sequence(root: etree, metadata: Iterable[object], series: str, num: int) -> None:
    """set series and series number metadata for the book"""
    remove_book_sequence(root, metadata)
    el = etree.SubElement(metadata, 'meta')
    el.attrib['name'] = META_CALIBRE_SEQ
    el.attrib['content'] = series
    el = etree.SubElement(metadata, 'meta')
    el.attrib['name'] = META_CALIBRE_INX
    el.attrib['content'] = str(num)



from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()


setup(
    name="epub-editor",
    packages=['epub_metadata'],
    package_dir={'epub_metadata': 'epub_metadata'},
    entry_points={
        'console_scripts': ['epub-editor=epub_editor:main'],
    },
    install_requires=[
        'lxml',
        'transliterate',
    ],
    version="0.0.3",
    url="https://github.com/shuvava/epub-metadata-editor",
    description="EPUB files metadata editor",
    license="GPL-3.0",
    keywords="epub editor",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",  # https://pypi.org/classifiers/
        "Environment :: Console",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
    ],
)
